from __future__ import annotations

import functools
import typing
import warnings

import walkman

__all__ = ("Cue", "CueManager")

ModuleNameToInitialiseKwargsDict = typing.Dict[str, typing.Union[bool, dict]]


class Cue(walkman.PlayMixin, walkman.JumpToMixin):
    def __init__(
        self,
        module_container: walkman.ModuleContainer,
        name: str,
        **kwargs: typing.Dict[int, typing.Union[dict, bool]],
    ):
        self._module_container = module_container
        self._name = name
        self._module_name_to_replication_configuration = kwargs
        # We can't call '_set_duration' when initialising a Cue,
        # because the cue won't be activated yet and the modules
        # will return wrong values for 'duration'.
        self._duration = None

    def _set_duration(self):
        if self._duration is None:
            duration_list = []
            for module in self.active_module_tuple:
                duration_list.append(module.duration)
            try:
                duration = max(duration_list)
            except ValueError:
                duration = 0
            self._duration = duration

    def _initialise_module_tuple(
        self,
        module_dict: typing.Dict[str, walkman.Module],
        replication_configuration: typing.Dict[int, typing.Dict[str, typing.Any]],
    ) -> typing.Tuple[walkman.Module, ...]:
        initialised_module_list = []
        for (
            module_replication_key,
            initialise_kwargs,
        ) in replication_configuration.items():
            try:
                module = module_dict[module_replication_key]
            except IndexError:
                warnings.warn(f"Replication '{module_replication_key}' doesn't exist.")
            else:
                try:
                    initialised_module_list.extend(
                        module.initialise(**initialise_kwargs)
                    )
                # Instead of providing kwargs, we can also simply
                # write 'false', to stop modules with 'auto_stop = false'
                except TypeError:
                    if initialise_kwargs is False:
                        module.stop()

        return tuple(initialised_module_list)

    def _play(self, duration: float = 0, delay: float = 0):
        for module in self.active_module_tuple:
            module.play(duration=duration, delay=delay)

    def _stop(
        self,
        wait: float = 0,
        module_to_keep_playing_tuple: typing.Tuple[walkman.Module, ...] = tuple([]),
        apply_auto_stop: bool = False,
    ):
        def stop_module(module_to_stop: walkman.Module, wait: float = 0):
            if module_to_stop not in module_to_keep_playing_tuple:
                # Auto stop
                if not apply_auto_stop or (
                    apply_auto_stop and module_to_stop.auto_stop
                ):
                    module_to_stop.stop(wait=wait)

        for main_module in self.active_main_module_tuple:
            stop_module(main_module, wait=wait)

        for (
            dependency_module,
            main_module_tuple,
        ) in self.dependency_module_to_main_module_tuple_dict.items():
            local_wait = (
                max(main_module.fade_out_duration for main_module in main_module_tuple)
                + wait
            )
            stop_module(
                dependency_module,
                wait=local_wait,
            )

    @functools.cached_property
    def active_main_module_tuple(self) -> typing.Tuple[walkman.Module, ...]:
        active_module_list = []
        for (
            module_name,
            replication_configuration,
        ) in self.module_name_to_replication_configuration.items():
            module_dict = self._module_container[module_name]
            for replication_key in replication_configuration.keys():
                active_module_list.append(module_dict[replication_key])
        return tuple(active_module_list)

    @functools.cached_property
    def main_module_to_dependency_module_chain_dict(
        self,
    ) -> typing.Dict[walkman.Module, tuple[walkman.Module]]:
        main_module_to_dependency_module_chain_dict = {}
        for main_module in self.active_main_module_tuple:
            dependency_module_chain = main_module.module_chain
            main_module_to_dependency_module_chain_dict.update(
                {main_module: dependency_module_chain}
            )
        return main_module_to_dependency_module_chain_dict

    @functools.cached_property
    def dependency_module_to_main_module_tuple_dict(
        self,
    ) -> typing.Dict[walkman.Module, tuple[walkman.Module]]:
        dependency_module_to_main_module_tuple_dict = {}
        for dependency_module in self.active_dependency_module_tuple:
            main_module_list = []
            for (
                main_module,
                dependency_module_chain,
            ) in self.main_module_to_dependency_module_chain_dict.items():
                if dependency_module in dependency_module_chain:
                    main_module_list.append(main_module)
            dependency_module_to_main_module_tuple_dict.update(
                {dependency_module: tuple(main_module_list)}
            )
        return dependency_module_to_main_module_tuple_dict

    @functools.cached_property
    def active_dependency_module_tuple(self) -> typing.Tuple[walkman.Module, ...]:
        active_dependency_module_list = []
        for (
            dependency_module_chain
        ) in self.main_module_to_dependency_module_chain_dict.values():
            for module in dependency_module_chain:
                if module not in active_dependency_module_list:
                    active_dependency_module_list.append(module)
        return tuple(active_dependency_module_list)

    @functools.cached_property
    def active_module_tuple(self) -> typing.Tuple[walkman.Module, ...]:
        active_module_list = []
        for module in (
            self.active_dependency_module_tuple + self.active_main_module_tuple
        ):
            if module not in active_module_list:
                active_module_list.append(module)
        return tuple(active_module_list)

    @functools.cached_property
    def name(self) -> str:
        return self._name

    @property
    def duration(self) -> float:
        if self._duration:
            return self._duration
        return 0

    @property
    def module_name_to_replication_configuration(
        self,
    ) -> ModuleNameToInitialiseKwargsDict:
        return self._module_name_to_replication_configuration

    def activate(self):
        initialised_module_list = []
        for (
            module_name,
            replication_configuration,
        ) in self.module_name_to_replication_configuration.items():
            try:
                module_dict = self._module_container[module_name]
            except KeyError:
                warnings.warn(f"Found invalid module_name '{module_name}'!")
            else:
                initialised_module_list.extend(
                    self._initialise_module_tuple(
                        module_dict, replication_configuration
                    )
                )

        self._set_duration()

        # Finally initialise all modules which didn't receive any
        # explicit settings.
        # XXX: We have to put parameters after other modules, to ensure
        # syntactic sugar is effective.
        for module_instance in sorted(
            self.active_module_tuple,
            key=lambda module: isinstance(module, walkman.Parameter),
        ):
            if module_instance not in initialised_module_list:
                initialised_module_list.extend(module_instance.initialise())

    def deactivate(
        self, module_to_keep_playing_tuple: typing.Tuple[walkman.Module, ...]
    ):
        if self.is_playing:
            self._is_playing = False
            self._stop(
                module_to_keep_playing_tuple=module_to_keep_playing_tuple,
                apply_auto_stop=True,
            )
        return self

    def jump_to(self, time_in_seconds: float):
        for module in self.active_module_tuple:
            module.jump_to(time_in_seconds)


class CueManager(object):
    def __init__(self, cue_tuple: typing.Tuple[Cue, ...]):
        self._cue_index = 0
        self._cue_tuple = cue_tuple
        self._cue_count = len(cue_tuple)

        self._active_cue_index = None
        self._activate_cue()

        self._cue_name_tuple = tuple(cue.name for cue in self._cue_tuple)

    def _set_cue_index(self, index: int):
        self._cue_index = index % self._cue_count

    def _activate_cue(self):
        if self._active_cue_index != self.current_cue_index:
            self.current_cue.activate()
            self._active_cue_index = self.current_cue_index
        self.current_cue.jump_to(0)

    def _move_and_activate_cue(self, cue_index: int):
        cue_to_deactivate = self.current_cue
        is_playing = cue_to_deactivate.is_playing
        self._set_cue_index(cue_index)
        cue_to_deactivate.deactivate(self.current_cue.active_module_tuple)
        self._activate_cue()
        if is_playing:
            self.current_cue.play()

    def __getitem__(self, index_or_key: typing.Union[str, int]) -> Cue:
        if isinstance(index_or_key, str):
            index = self._cue_name_tuple.index(index_or_key)
        else:
            index = index_or_key
        return self._cue_tuple[index]

    @property
    def cue_name_tuple(self) -> typing.Tuple[str, ...]:
        return self._cue_name_tuple

    @property
    def active_module_tuple(self):
        return self.current_cue.active_module_tuple

    @property
    def cue_count(self):
        return self._cue_count

    @property
    def current_cue_index(self):
        return self._cue_index

    @property
    def current_cue(self) -> Cue:
        return self._cue_tuple[self._cue_index]

    def move_to_previous_cue(self):
        self._move_and_activate_cue(self._cue_index - 1)

    def move_to_next_cue(self):
        self._move_and_activate_cue(self._cue_index + 1)

    def jump_to_cue(self, cue_index: int):
        self._move_and_activate_cue(cue_index)



from .version import __version__
from . import constants
from . import utilities
from .mixins import *
from .channel_mappings import *
from .audio import *
from .modules import *
from .cues import *
from . import tests
from .backends import Backend
from . import parsers
from . import ui
from .boot import *
from . import unit_tests



import typing

import pyo


__all__ = ("ChannelMapping", "dict_or_channel_mapping_to_channel_mapping")


class ChannelMapping(typing.Dict[int, typing.Tuple[int, ...]]):
    """Map audio channels to other audio channels.

    Left side is always one channel, right side can be one
    or more channels.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Hack to ensure key/values are integer
        key_to_remove_list = []
        update_dict = {}
        for key, value in self.items():
            if not isinstance(key, int):
                key_to_remove_list.append(key)
            if isinstance(value, (str, int)):
                output_channel_list = [value]
            else:
                output_channel_list = value
            output_channel_tuple = tuple(
                int(output_channel) for output_channel in output_channel_list
            )
            update_dict[int(key)] = output_channel_tuple
        self.update(update_dict)
        for key in key_to_remove_list:
            del self[key]

    @property
    def left_channel_set(self) -> typing.Set[int]:
        return set(self.keys())

    @property
    def maxima_left_channel(self) -> int:
        return max(self.left_channel_set) + 1

    @property
    def right_channel_tuple_tuple(self) -> typing.Tuple[typing.Tuple[int, ...], ...]:
        return tuple(self.values())

    @property
    def right_channel_set(self) -> typing.Set[int]:
        right_channel_set = set([])
        for right_channel_tuple in self.right_channel_tuple_tuple:
            for right_channel in right_channel_tuple:
                right_channel_set.add(right_channel)
        return right_channel_set

    @property
    def maxima_right_channel(self) -> int:
        return max(self.right_channel_set) + 1

    def to_mixer(self) -> pyo.Mixer:
        return pyo.Mixer(outs=self.maxima_right_channel)


def dict_or_channel_mapping_to_channel_mapping(
    dict_or_channel_mapping_to_channel_mapping: typing.Union[dict, ChannelMapping]
) -> ChannelMapping:
    if not isinstance(dict_or_channel_mapping_to_channel_mapping, ChannelMapping):
        channel_mapping = ChannelMapping(dict_or_channel_mapping_to_channel_mapping)
    else:
        channel_mapping = dict_or_channel_mapping_to_channel_mapping
    return channel_mapping



from __future__ import annotations

import functools
import typing
import warnings

import pyo

import walkman

from . import base

__all__ = (
    "ModuleWithUneffectiveStop",
    "Empty",
    "Parameter",
    "AudioInput",
    "MidiControlInput",
    "ModuleWithDecibel",
    "Sine",
    "Amplification",
    "Mixer",
    "Filter",
    "ConvolutionReverb",
    "WaveguideReverb",
    "ButterworthLowpassFilter",
    "ButterworthHighpassFilter",
    "Equalizer",
)


class ModuleWithUneffectiveStop(base.Module):
    def _stop(self, _: float = 0):
        ...

    def stop(self, _: float = 0) -> ModuleWithUneffectiveStop:
        return self


class Empty(base.Module):
    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.signal = pyo.Sig(0)
        self.internal_pyo_object_list.append(self.signal)

    @property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.signal


class IllegalMidiControlIndexWarning(Warning):
    pass


DEFAULT_ENVELOPE_TYPE = "linear"


class UndefinedEnvelopeTypeWarning(Warning):
    def __init__(self, undefined_envelope_type: str):
        super().__init__(
            f"Found undefined envelope type '{undefined_envelope_type}'. "
            f"WALKMAN set envelope type to {DEFAULT_ENVELOPE_TYPE}."
        )


class Parameter(base.Module):
    ITEM_TO_VOICE_DICT = {
        "VALUE_VOICE": 0,
        "LINEAR_ENVELOPE_VOICE": 1,
        "EXPONENTIAL_ENVELOPE_VOICE": 2,
    }
    ENVELOPE_TYPE_TO_VOICE = {
        "linear": ITEM_TO_VOICE_DICT["LINEAR_ENVELOPE_VOICE"],
        "exponential": ITEM_TO_VOICE_DICT["EXPONENTIAL_ENVELOPE_VOICE"],
    }

    def __init__(
        self,
        fade_in_duration: float = 0.0001,
        fade_out_duration: float = 0.00001,
        **kwargs,
    ):
        super().__init__(
            fade_in_duration=fade_in_duration,
            fade_out_duration=fade_out_duration,
            **kwargs,
        )

    def _setup_pyo_object(
        self,
    ):
        super()._setup_pyo_object()

        # If value is only a numerical value, we will only
        # pass it to 'pyo.Sig'
        self.signal = pyo.Sig(0)
        # Scale is used if a midi control input is used.
        # It get initialised with dummy values
        self.linear_envelope = pyo.Linseg([(0, 0), (2, 0)])
        self.exponential_envelope = pyo.Expseg([(0, 0.1), (2, 0.1)])
        self.selector = pyo.Selector(
            [
                self.signal,
                self.linear_envelope,
                self.exponential_envelope,
            ]
        )
        self.portamento = pyo.Port(self.selector)

        self.ENVELOPE_TYPE_TO_ENVELOPE = {
            "linear": self.linear_envelope,
            "exponential": self.exponential_envelope,
        }

        self.internal_pyo_object_list.extend(
            [
                self.signal,
                self.linear_envelope,
                self.exponential_envelope,
                self.selector,
                self.portamento,
            ]
        )

        self.envelope_list = None
        self.envelope = None
        self.start_time: float = 0
        self.time: float = 0

    @property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.portamento

    @property
    def duration(self) -> float:
        return self.envelope_list[-1][0] if self.envelope_list else 0

    def _initialise(
        self,
        value: typing.Union[float, typing.List[typing.List[float]]] = 0,
        envelope_type: str = DEFAULT_ENVELOPE_TYPE,
        rise_time: float = 0.001,
        fall_time: float = 0.001,
    ):
        self.portamento.setRiseTime(rise_time)
        self.portamento.setFallTime(fall_time)
        selector_voice = None
        envelope_list = None
        if isinstance(value, list):
            envelope_list = list(tuple(point) for point in value)
            try:
                envelope = self.ENVELOPE_TYPE_TO_ENVELOPE[envelope_type]
            except KeyError:
                envelope_type = DEFAULT_ENVELOPE_TYPE
                warnings.warn(envelope_type, UndefinedEnvelopeTypeWarning)
            finally:
                envelope = self.ENVELOPE_TYPE_TO_ENVELOPE[envelope_type]

            envelope.setList(envelope_list)
            selector_voice = self.ENVELOPE_TYPE_TO_VOICE[envelope_type]
            self.envelope = envelope

        else:
            self.signal.setValue(value)
            selector_voice = self.ITEM_TO_VOICE_DICT["VALUE_VOICE"]

        self.envelope_list = envelope_list
        self.selector.setVoice(selector_voice)
        self.time = 0


class ModuleWithDecibel(base.Module, decibel=base.AutoSetup(Parameter)):
    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.amplitude = pyo.DBToA(self.decibel.pyo_object)
        self.amplitude_signal_to = pyo.SigTo(self.amplitude, time=0.015)
        self.internal_pyo_object_list.extend([self.amplitude, self.amplitude_signal_to])


class AudioInput(ModuleWithDecibel):
    def __init__(self, input_channel_index: int = 0, **kwargs):
        super().__init__(**kwargs)
        self.input_channel_index = input_channel_index

    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.audio_input = pyo.Input(self.input_channel_index)
        self.denorm = pyo.Denorm(self.audio_input, mul=self.amplitude_signal_to)
        self.internal_pyo_object_list.extend([self.audio_input, self.denorm])

    @property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.denorm


class MidiControlInput(ModuleWithUneffectiveStop):
    def __init__(self, midi_control_number: int = 0, midi_channel: int = 0, **kwargs):
        super().__init__(**kwargs)
        self.midi_control_number = midi_control_number
        self.midi_channel = midi_channel

    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.midi_input = pyo.Midictl(
            self.midi_control_number, channel=self.midi_channel
        )
        self.internal_pyo_object_list.append(self.midi_input)

    def _initialise(self, minima: float = 0, maxima: float = 1, **_):
        self.midi_input.setMinScale(minima)
        self.midi_input.setMaxScale(maxima)

    @property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.midi_input


class Sine(
    ModuleWithDecibel,
    frequency=base.AutoSetup(
        Parameter, module_kwargs={"default_dict": {"value": 1000}}
    ),
):
    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.sine = pyo.Sine(
            freq=self.frequency.pyo_object, mul=self.amplitude_signal_to
        )
        self.internal_pyo_object_list.append(self.sine)

    @property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.sine


class Amplification(
    ModuleWithDecibel,
    audio_input=base.Catch(walkman.constants.EMPTY_MODULE_INSTANCE_NAME),
):
    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.amplification = self.amplitude_signal_to * self.audio_input.pyo_object
        self.internal_pyo_object_list.append(self.amplification)

    @functools.cached_property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.amplification


class IllegalChannelIndexWarning(Warning):
    """Call if user gave too high channel"""


MixerIndex = int
MixerInfo = typing.Tuple[MixerIndex, ...]
"""Each channel of an audio signal has a specific index in the mixer"""


class Mixer(
    ModuleWithDecibel,
    **{
        f"audio_input_{index}": base.Catch(
            walkman.constants.EMPTY_MODULE_INSTANCE_NAME, relevance=False
        )
        for index in range(100)
    },
):
    class PyoObjectMixer(pyo.PyoObject):
        def __init__(self, base_objs: list):
            super().__init__()
            self._base_objs = base_objs

    def __init__(
        self,
        channel_mapping: typing.Union[walkman.ChannelMapping, dict] = {0: 0, 1: 1},
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.channel_mapping = walkman.dict_or_channel_mapping_to_channel_mapping(
            channel_mapping
        )
        self.input_channel_count = self.channel_mapping.maxima_left_channel
        self.output_channel_count = self.channel_mapping.maxima_right_channel
        self.audio_input_key_to_mixer_info = {}

    def _setup_pyo_object(self):
        super()._setup_pyo_object()

        self.input_mixer = pyo.Mixer(outs=self.input_channel_count)
        self.output_mixer = pyo.Mixer(
            outs=self.output_channel_count  # , mul=self.amplitude_signal_to
        )

        for (
            input_channel_index,
            output_channel_index_tuple,
        ) in self.channel_mapping.items():
            local_pyo_object = self.PyoObjectMixer(
                self.input_mixer[input_channel_index]
            )
            self.output_mixer.addInput(input_channel_index, local_pyo_object)
            for output_channel_index in output_channel_index_tuple:
                self.output_mixer.setAmp(input_channel_index, output_channel_index, 1)

            self.internal_pyo_object_list.append(local_pyo_object)

        self.output_mixer_as_pyo_object = self.PyoObjectMixer(
            [channel[0] for channel in self.output_mixer]
        )

        self.internal_pyo_object_list.extend(
            [self.input_mixer, self.output_mixer, self.output_mixer_as_pyo_object]
        )

    @functools.cached_property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.output_mixer_as_pyo_object

    def _initialise(
        self, **audio_input_channel_mapping: typing.Dict[str, walkman.ChannelMapping]
    ):
        audio_input_key_to_channel_mapping_dict = {}
        for (
            audio_input_channel_mapping_key,
            channel_mapping,
        ) in audio_input_channel_mapping.items():
            audio_input_key, *_ = audio_input_channel_mapping_key.split(
                "_channel_mapping"
            )
            if audio_input_key in self.module_input_dict.keys():
                audio_input_key_to_channel_mapping_dict.update(
                    {audio_input_key: channel_mapping}
                )
            else:
                warnings.warn(
                    "Found invalid channel mapping "
                    f"argument '{audio_input_channel_mapping_key}'!"
                )

        for audio_input_key in self.module_input_dict.keys():
            module_instance = getattr(self, audio_input_key)
            if not isinstance(module_instance, Empty):
                if audio_input_key not in audio_input_key_to_channel_mapping_dict:
                    channel_mapping = {}
                    for channel_index in range(len(module_instance.pyo_object)):
                        output_channel = channel_index % self.input_channel_count
                        channel_mapping.update({channel_index: output_channel})
                    audio_input_key_to_channel_mapping_dict.update(
                        {audio_input_key: channel_mapping}
                    )

        for (
            audio_input_key,
            channel_mapping,
        ) in audio_input_key_to_channel_mapping_dict.items():
            if isinstance(channel_mapping, (dict, walkman.ChannelMapping)):
                channel_mapping = walkman.dict_or_channel_mapping_to_channel_mapping(
                    channel_mapping
                )
                self.activate_channel_mapping(audio_input_key, channel_mapping)

    def _raise_illegal_audio_object_channel_warning(
        self,
        audio_input_key: str,
        channel_mapping: walkman.ChannelMapping,
        audio_object_channel_index: int,
    ):
        warnings.warn(
            "Found invalid audio_object_channel_index ="
            f" {audio_object_channel_index} in channel_mapping ="
            f" {channel_mapping} for audio input key '{audio_input_key}'.",
            IllegalChannelIndexWarning,
        )

    def register_audio_object(self, audio_input_key: str) -> typing.Optional[MixerInfo]:
        mixer_info_list = []
        base_index = walkman.utilities.get_next_mixer_index(self.input_mixer)
        try:
            pyo_object = getattr(self, audio_input_key).pyo_object
        except AttributeError:
            warnings.warn(
                f"No audio input with audio input key '{audio_input_key}' exists!"
                " WALKMAN ignored settings for invalid audio input key."
            )
            return None
        for audio_object_index in range(len(pyo_object)):
            pyo_channel = pyo_object[audio_object_index]
            mixer_index = base_index + audio_object_index
            self.input_mixer.addInput(mixer_index, pyo_channel)
            mixer_info_list.append(mixer_index)

        mixer_info = tuple(mixer_info_list)
        self.audio_input_key_to_mixer_info.update({audio_input_key: mixer_info})
        return mixer_info

    def activate_channel_mapping(
        self,
        audio_input_key: str,
        channel_mapping: walkman.ChannelMapping,
    ):
        try:
            mixer_info = self.audio_input_key_to_mixer_info[audio_input_key]
        except KeyError:
            mixer_info = self.register_audio_object(audio_input_key)
            if mixer_info is None:
                return

        for (
            audio_object_channel_index,
            output_channel_index_tuple,
        ) in channel_mapping.items():
            try:
                mixer_index = mixer_info[audio_object_channel_index]
            except IndexError:
                self._raise_illegal_audio_object_channel_warning(
                    audio_input_key, channel_mapping, audio_object_channel_index
                )
            else:
                for module_output_channel_index in range(self.input_channel_count):
                    amplitude = int(
                        module_output_channel_index in output_channel_index_tuple
                    )
                    self.input_mixer.setAmp(
                        mixer_index, module_output_channel_index, amplitude
                    )


class Filter(
    ModuleWithDecibel,
    frequency=base.AutoSetup(
        Parameter, module_kwargs={"default_dict": {"value": 1000}}
    ),
    q=base.AutoSetup(Parameter, module_kwargs={"default_dict": {"value": 5}}),
    audio_input=base.Catch(walkman.constants.EMPTY_MODULE_INSTANCE_NAME),
):
    FILTER_TYPE_TO_INTERNAL_FILTER_TYPE = {
        "lowpass": 0,
        "highpass": 1,
        "bandpass": 2,
        "bandstop": 3,
        "allpass": 4,
    }

    def __init__(self, stages: int = 4, filter_type: str = "lowpass", **kwargs):
        super().__init__(**kwargs)
        self.stages = stages
        try:
            internal_filter_type = self.FILTER_TYPE_TO_INTERNAL_FILTER_TYPE[filter_type]
        except KeyError:
            default_filter_type = "lowpass"
            internal_filter_type = self.FILTER_TYPE_TO_INTERNAL_FILTER_TYPE[
                default_filter_type
            ]
            warnings.warn(
                f"Found undefined filter type '{filter_type}'."
                f"Filter type has been set to '{default_filter_type}'."
            )
        self.internal_filter_type = internal_filter_type

    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.audio_filter = pyo.Biquadx(
            self.audio_input.pyo_object,
            mul=self.amplitude_signal_to,
            freq=self.frequency.pyo_object,
            q=self.q.pyo_object,
            stages=self.stages,
            type=self.internal_filter_type,
        )
        self.internal_pyo_object_list.append(self.audio_filter)

    @functools.cached_property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.audio_filter


class ConvolutionReverb(
    ModuleWithDecibel,
    audio_input=base.Catch(walkman.constants.EMPTY_MODULE_INSTANCE_NAME),
    balance=base.AutoSetup(Parameter, module_kwargs={"default_dict": {"value": 1}}),
):
    def __init__(self, *, impulse_path: str, sample_size: int = 1024, **kwargs):
        super().__init__(**kwargs)
        self.impulse_path = impulse_path
        self.sample_size = sample_size

    def _setup_pyo_object(self):
        super()._setup_pyo_object()

        self.convolution_reverb = pyo.CvlVerb(
            self.audio_input.pyo_object,
            self.impulse_path,
            size=self.sample_size,
            mul=self.amplitude_signal_to,
            bal=self.balance.pyo_object,
        ).stop()

        self.internal_pyo_object_list.append(self.convolution_reverb)

    @functools.cached_property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.convolution_reverb


class WaveguideReverb(
    ModuleWithDecibel,
    audio_input=base.Catch(walkman.constants.EMPTY_MODULE_INSTANCE_NAME),
    balance=base.AutoSetup(Parameter, module_kwargs={"default_dict": {"value": 1}}),
    cutoff_frequency=base.AutoSetup(
        Parameter, module_kwargs={"default_dict": {"value": 6000}}
    ),
    feedback=base.AutoSetup(Parameter, module_kwargs={"default_dict": {"value": 0.6}}),
):
    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.waveguide_reverb = pyo.WGVerb(
            self.audio_input.pyo_object,
            feedback=self.feedback.pyo_object,
            mul=self.amplitude_signal_to,
            bal=self.balance.pyo_object,
        ).stop()
        self.internal_pyo_object_list.append(self.waveguide_reverb)

    @functools.cached_property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.waveguide_reverb


class ButterworthLowpassFilter(
    ModuleWithDecibel,
    audio_input=base.Catch(walkman.constants.EMPTY_MODULE_INSTANCE_NAME),
    frequency=base.AutoSetup(
        Parameter, module_kwargs={"default_dict": {"value": 1000}}
    ),
):
    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.lowpass_filter = pyo.ButLP(
            self.audio_input.pyo_object,
            mul=self.amplitude_signal_to,
            freq=self.frequency.pyo_object,
        ).stop()
        self.internal_pyo_object_list.append(self.lowpass_filter)

    @functools.cached_property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.lowpass_filter


class ButterworthHighpassFilter(
    ModuleWithDecibel,
    audio_input=base.Catch(walkman.constants.EMPTY_MODULE_INSTANCE_NAME),
    frequency=base.AutoSetup(
        Parameter, module_kwargs={"default_dict": {"value": 1000}}
    ),
):
    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.highpass_filter = pyo.ButHP(
            self.audio_input.pyo_object,
            mul=self.amplitude_signal_to,
            freq=self.frequency.pyo_object,
        ).stop()
        self.internal_pyo_object_list.append(self.highpass_filter)

    @functools.cached_property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.highpass_filter


class Equalizer(
    ModuleWithDecibel,
    audio_input=base.Catch(walkman.constants.EMPTY_MODULE_INSTANCE_NAME),
    frequency=base.AutoSetup(Parameter, module_kwargs={"default_dict": {"value": 500}}),
    boost=base.AutoSetup(Parameter, module_kwargs={"default_dict": {"value": -3}}),
):
    FILTER_TYPE_TO_INTERNAL_FILTER_TYPE = {
        "peak": 0,
        "notch": 0,
        "lowshelf": 1,
        "highshelf": 2,
    }

    def __init__(self, filter_type: str = "peak", **kwargs):
        super().__init__(**kwargs)
        try:
            internal_filter_type = self.FILTER_TYPE_TO_INTERNAL_FILTER_TYPE[filter_type]
        except KeyError:
            default_filter_type = "peak"
            warnings.warn(
                f"Found undefined filter type '{filter_type}'."
                f"Filter type has been set to '{default_filter_type}'."
            )
            internal_filter_type = self.FILTER_TYPE_TO_INTERNAL_FILTER_TYPE[
                default_filter_type
            ]
        self.internal_filter_type = internal_filter_type

    def _setup_pyo_object(self):
        super()._setup_pyo_object()
        self.equalizer = pyo.EQ(
            self.audio_input.pyo_object,
            mul=self.amplitude_signal_to,
            freq=self.frequency.pyo_object,
            boost=self.boost.pyo_object,
            type=self.internal_filter_type,
        ).stop()
        self.internal_pyo_object_list.append(self.equalizer)

    def _initialise(self, q: float = 1, **_):
        self.equalizer.setQ(q)

    @functools.cached_property
    def _pyo_object(self) -> pyo.PyoObject:
        return self.equalizer



from __future__ import annotations

import pyo

import walkman

__all__ = ("AudioHost",)


class AudioHost(walkman.PlayMixin, walkman.CloseMixin, walkman.DecibelMixin):
    """Wrapper for pyo.Server.

    Simplifies server API.
    """

    def __init__(
        self,
        audio: str = "jack",
        midi: str = "jack",
        sampling_rate: int = 44100,
        buffer_size: int = 1024,
        channel_count: int = 2,
        **kwargs,
    ):
        self.channel_count = channel_count
        self.server = pyo.Server(
            sr=sampling_rate,
            midi=midi,
            nchnls=channel_count,
            buffersize=buffer_size,
            duplex=1,
            audio=audio,
            jackname=f"{walkman.constants.NAME}_{channel_count}",
            **kwargs,
        )
        self.server.setJackAuto(xin=False, xout=False)
        self.server.boot()
        self.decibel = -12


    def _play(self, **_):
        self.server.start()

    def _stop(self, wait: float = 0):
        self.server.stop()

    def close(self):
        self.stop()
        del self.server

    @property
    def decibel(self) -> float:
        return self._decibel

    @decibel.setter
    def decibel(self, decibel: float):
        self._decibel = decibel
        self.server.setAmp(float(1 * (10 ** (decibel / 20))))



from __future__ import annotations

import abc
import copy
import dataclasses
import functools
import importlib
import inspect
import pkgutil
import typing

import pyo

import walkman


__all__ = ("ModuleInput", "Catch", "AutoSetup", "Module", "ModuleContainer")


def override_default_kwargs(method_to_wrap: typing.Callable) -> typing.Callable:
    def wrapper(self, **kwargs):
        default_dict = copy.deepcopy(self.default_dict)
        default_dict.update(kwargs)
        return method_to_wrap(self, **default_dict)

    return wrapper


class ModuleInput(abc.ABC):
    """Allocate an Module object as an input of another Module"""

    def __init__(self, relevance: bool = True):
        self.relevance = relevance

    @abc.abstractmethod
    def __call__(self, module_container: ModuleContainer) -> Module:
        ...


class Catch(ModuleInput):
    def __init__(self, module_instance_name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.module_instance_name = module_instance_name

    def __call__(self, module_container: ModuleContainer) -> Module:
        try:
            return module_container.get_module_by_name(self.module_instance_name)
        except (KeyError, InvalidModuleInstanceNameError):
            return module_container.get_module_by_name(
                walkman.constants.EMPTY_MODULE_INSTANCE_NAME
            )


class AutoSetup(ModuleInput):
    def __init__(
        self,
        module_class: typing.Type[Module],
        module_kwargs: dict = {},
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.module_class = module_class
        self.module_kwargs = module_kwargs

    def __call__(self, module_container: ModuleContainer) -> Module:
        module = self.module_class(**self.module_kwargs)
        module_instance_dict = {id(module): module}
        module_name = module.get_class_name()
        try:
            module_container[module_name].update(module_instance_dict)
        except KeyError:
            module_container.update({module_name: module_instance_dict})
        return module


@dataclasses.dataclass()
class Module(
    walkman.PlayMixin,
    walkman.JumpToMixin,
    walkman.NamedMixin,
    walkman.CloseMixin,
    walkman.PyoObjectMixin,
):
    send_to_physical_output: bool = False
    auto_stop: bool = True
    fade_in_duration: float = 0.001
    fade_out_duration: float = 0.001
    output_module_set: typing.Set[Module] = dataclasses.field(
        default_factory=lambda: set([])
    )
    module_input_dict: typing.Dict[str, ModuleInput] = dataclasses.field(
        default_factory=lambda: {}
    )
    default_dict: typing.Dict[str, typing.Any] = dataclasses.field(
        default_factory=lambda: {}
    )
    internal_pyo_object_list: typing.List[pyo.PyoObject] = dataclasses.field(
        default_factory=lambda: []
    )

    def __init_subclass__(cls, **module_input_dict: ModuleInput):
        try:
            cls.default_module_input_dict = dict(
                cls.default_module_input_dict, **module_input_dict
            )
        except AttributeError:
            cls.default_module_input_dict = module_input_dict
            is_subclass = False
        else:
            is_subclass = True

        if not is_subclass:
            user_defined_init = cls.__init__

            def __init__(self, **kwargs):
                module_input_dict = {}
                for (
                    module_input_name,
                    default_module_input,
                ) in self.default_module_input_dict.items():
                    module_input = default_module_input
                    try:
                        module_instance_name = kwargs[module_input_name]
                    except KeyError:
                        pass
                    else:
                        module_input = Catch(
                            module_instance_name, relevance=module_input.relevance
                        )
                        del kwargs[module_input_name]
                    finally:
                        module_input_dict.update({module_input_name: module_input})

                user_defined_init(self, module_input_dict=module_input_dict, **kwargs)

            cls.__init__ = __init__

    def __hash__(self) -> int:
        # XXX: This method is overridden by dataclasses and
        # has to be manually added again.
        return walkman.NamedMixin.__hash__(self)

    # ################## ABSTRACT PROPERTIES ################## #

    @abc.abstractproperty
    def _pyo_object(self) -> pyo.PyoObject:
        ...

    # ################## ABSTRACT METHODS    ################## #

    def _initialise(self, **_):
        ...

    # ################## PRIVATE METHODS     ################## #

    def _play(self, duration: float = 0, delay: float = 0):
        # self.fader.play(dur=duration, delay=delay)
        self.fader.play(dur=duration, delay=delay)
        self.pyo_object.play(dur=duration, delay=delay)
        for internal_pyo_object in self.internal_pyo_object_list:
            internal_pyo_object.play(dur=duration, delay=delay)

        if self.send_to_physical_output:
            for channel_index, audio_stream in enumerate(self.pyo_object):
                audio_stream.out(channel_index)

    def _stop(self, wait: float = 0):
        self.fader_stopper.play(delay=wait)
        wait += self.fade_out_duration
        for internal_pyo_object in self.internal_pyo_object_list:
            internal_pyo_object.stop(wait=wait)
        self.pyo_object.stop(wait=wait)

    # ################## PUBLIC METHODS      ################## #

    def setup(self, module_container: ModuleContainer):
        self.assign_module_inputs(module_container)
        self.setup_pyo_object()

    def assign_module_inputs(self, module_container: ModuleContainer):
        for module_input_name, module_input in self.module_input_dict.items():
            module = module_input(module_container)
            module.output_module_set.add(self)
            setattr(self, module_input_name, module)

    def setup_pyo_object(self):
        self.fader = pyo.Fader(
            fadein=self.fade_in_duration, fadeout=self.fade_out_duration
        )
        # XXX: We need to use a trigger for stopping the fader, because
        # fader objects ignore the 'wait' parameter. Without this trigger
        # the wait parameter wouldn't work.
        self.fader_stopper = pyo.Trig().stop()
        self.fader_stopper_function = pyo.TrigFunc(
            input=self.fader_stopper, function=self.fader.stop
        ).play()

    @override_default_kwargs
    def initialise(
        self,
        **kwargs,
    ):
        print("INITIALISE")
        # Parse everything else to actual initialise method
        self._initialise(**kwargs)

    # ################## PUBLIC PROPERTIES  ################## #

    @property
    def duration(self) -> float:
        return 0

    @functools.cached_property
    def pyo_object(self) -> pyo.PyoObject:
        pyo_object = self._pyo_object * self.fader
        pyo_object.stop()
        return pyo_object

    @functools.cached_property
    def module_input_chain(self) -> typing.Tuple[Module, ...]:
        """Get chain of inputs from left to right (order matters!)"""

        input_module_list = []
        for module_input_name in self.module_input_dict.keys():
            input_module = getattr(self, module_input_name)
            input_module_list.append(input_module)
            input_module_list.extend(input_module.module_input_chain)
        return tuple(reversed(input_module_list))

    @functools.cached_property
    def relevant_module_input_chain(self) -> typing.Tuple[Module, ...]:
        """Get chain of relevant inputs from left to right (order matters!)"""

        input_module_list = []
        for module_input_name, module_input in self.module_input_dict.items():
            if module_input.relevance:
                input_module = getattr(self, module_input_name)
                if input_module not in input_module_list:
                    input_module_list.append(input_module)
                for module_instance in input_module.relevant_module_input_chain:
                    if module_instance not in input_module_list:
                        input_module_list.append(module_instance)
        return tuple(reversed(input_module_list))

    @functools.cached_property
    def module_output_chain(self) -> typing.Tuple[Module, ...]:
        output_module_list = []
        for output_module in self.output_module_set:
            if output_module not in output_module_list:
                output_module_list.append(output_module)
            for module_instance in output_module.relevant_module_chain:
                if (
                    module_instance not in output_module_list
                    and module_instance != self
                ):
                    output_module_list.append(module_instance)
        return tuple(output_module_list)

    @functools.cached_property
    def module_chain(self) -> typing.Tuple[Module, ...]:
        module_list = list(self.module_input_chain)
        for module_instance in self.module_output_chain:
            if module_instance not in module_list:
                module_list.append(module_instance)
        return tuple(module_list)

    @functools.cached_property
    def relevant_module_chain(self) -> typing.Tuple[Module, ...]:
        module_list = list(self.relevant_module_input_chain)
        for module_instance in self.module_output_chain:
            if module_instance not in module_list:
                module_list.append(module_instance)
        return tuple(module_list)


class InvalidModuleInstanceNameError(Exception):
    def __init__(self, invalid_module_instance_name: str):
        super().__init__(
            f"Found invalid module instance name '{invalid_module_instance_name}'!"
        )


ModuleName = str
ReplicationKey = str
ModuleNameToModuleClassDict = typing.Dict[str, typing.Type[Module]]


class ModuleContainer(
    typing.Dict[ModuleName, typing.Dict[ReplicationKey, Module]], walkman.CloseMixin
):
    @staticmethod
    def get_module_name_to_module_class_dict() -> ModuleNameToModuleClassDict:
        namespace_package = __import__(walkman.constants.MODULE_PACKAGE_NAME)
        module_list = list(
            importlib.import_module(module_info.name)
            for module_info in pkgutil.iter_modules(
                namespace_package.__path__, namespace_package.__name__ + "."
            )
        )
        module_list.append(importlib.import_module("walkman.modules.buildins"))
        module_name_to_module_class_dict = {}
        for module in module_list:
            for _, python_class in inspect.getmembers(module, inspect.isclass):
                if Module in inspect.getmro(python_class):
                    module_name_to_module_class_dict.update(
                        {python_class.get_name(): python_class}
                    )
        return module_name_to_module_class_dict

    @classmethod
    def from_module_configuration(
        cls,
        module_name_to_replication_configuration_dict: typing.Dict[
            str, typing.Dict[str, typing.Dict[str, typing.Any]]
        ],
        module_name_to_module_class_dict: typing.Optional[
            ModuleNameToModuleClassDict
        ] = None,
    ) -> ModuleContainer:
        if module_name_to_module_class_dict is None:
            module_name_to_module_class_dict = (
                cls.get_module_name_to_module_class_dict()
            )

        module_name_to_module_container = {}
        for module_name, module_class in module_name_to_module_class_dict.items():
            try:
                replication_key_to_module_configuration_dict = (
                    module_name_to_replication_configuration_dict[module_name]
                )
            except KeyError:
                replication_key_to_module_configuration_dict = {}
            module_dict = {}
            for (
                replication_key,
                module_configuration,
            ) in replication_key_to_module_configuration_dict.items():
                module = module_class(**module_configuration)
                module_dict[replication_key] = module
            module_name_to_module_container.update({module_name: module_dict})
        return cls(module_name_to_module_container)

    def close(self):
        for module in self.module_tuple:
            module.close()

    @property
    def module_tuple(self) -> typing.Tuple[walkman.Module, ...]:
        module_list: typing.List[walkman.Module] = []
        for module_tuple in self.values():
            module_list.extend(module_tuple)
        return tuple(module_list)

    def get_module_by_name(self, module_instance_name: str) -> Module:
        try:
            module_name, replication_key = module_instance_name.split(".")
        except ValueError:
            raise InvalidModuleInstanceNameError(module_instance_name)
        else:
            return self[module_name][replication_key]



import dataclasses
import functools
import itertools
import typing
import uuid

import pyo

import walkman


@dataclasses.dataclass
class AudioTest(object):
    channel_count: int
    internal_pyo_object_list: typing.List[pyo.PyoObject] = dataclasses.field(
        default_factory=lambda: []
    )

    @functools.cached_property
    def name(self) -> str:
        return str(uuid.uuid4())

    def get_name(self):
        return self.name

    def close(self):
        # Remove any test traces
        for internal_pyo_object in self.internal_pyo_object_list:
            internal_pyo_object.stop()
            del internal_pyo_object


@dataclasses.dataclass
class AudioRotationTest(AudioTest):
    play_duration: float = 1

    def __post_init__(self):
        self._channel_index_cycle = itertools.cycle(range(self.channel_count))
        self._decibel = -120
        self.decibel_signal = pyo.SigTo(self._decibel)
        self.decibel_to_amplitude = pyo.DBToA(self.decibel_signal)
        self.fader = pyo.Fader()
        self.noise = pyo.Noise(mul=self.fader * self.decibel_to_amplitude)
        self.mixer = pyo.Mixer(outs=self.channel_count)
        self.mixer.addInput(0, self.noise)
        self.pattern = pyo.Pattern(self._increment, self.play_duration + 0.075)
        self.pattern.stop()
        self.internal_pyo_object_list.extend(
            [
                self.pattern,
                self.mixer,
                self.noise,
                self.fader,
                self.decibel_signal,
                self.decibel_to_amplitude,
            ]
        )
        for output_index, channel in enumerate(self.mixer):
            channel[0].out(output_index)

    def _increment(self):
        channel_index = next(self._channel_index_cycle)
        walkman.constants.LOGGER.info(
            f"{type(self).__name__}: Test channel {channel_index}."
        )
        for output_channel_index in range(self.channel_count):
            amplitude = output_channel_index == channel_index
            self.mixer.setAmp(0, output_channel_index, amplitude)
        self.fader.play(dur=self.play_duration)

    @property
    def decibel(self) -> float:
        return self._decibel

    @decibel.setter
    def decibel(self, decibel: float):
        self.decibel_signal.setValue(decibel)

    @property
    def pyo_object_list(self) -> typing.List[pyo.PyoObject]:
        return self.internal_pyo_object_list

    def play(self):
        self.pattern.play()
        self.noise.play()

    def stop(self):
        self.pattern.stop()
        self.noise.stop()



import abc
import math
import inspect
import typing
import unittest

import pyo

import walkman


class SineAudioObject(walkman.PyoObjectMixin):
    """Minimalistic audio object useful to unit testing."""

    def __init__(self, *args, **kwargs):
        self._pyo_object = pyo.Sine(*args, **kwargs)

    @property
    def pyo_object(self) -> pyo.PyoObject:
        return self._pyo_object


class WalkmanTestCase(unittest.TestCase):
    server_args: typing.Tuple[typing.Any, ...] = tuple([])
    # XXX: buffer size is set to an extremely low value, to
    # ensure tests give precise results. If we use
    # a higher buffer size it is more likely that we get
    # rounding errors.
    server_kwargs: typing.Dict[str, typing.Any] = {"audio": "manual", "buffersize": 8}

    def process_server(self, buffer_count: int):
        for _ in range(buffer_count):
            self.server.process()

    def jump_to(self, duration_in_seconds: float):
        buffer_count = int(
            math.ceil(
                duration_in_seconds
                * self.server.getSamplingRate()
                / self.server.getBufferSize()
            )
        )
        self.process_server(buffer_count)

    def is_pyo_object_not_silent(
        self, pyo_object_to_test: pyo.PyoObject, duration_in_seconds: float = 2
    ) -> bool:
        is_active_list = []
        is_silent_tracker = pyo.Change(pyo_object_to_test)
        is_silent_writer = pyo.TrigFunc(
            is_silent_tracker, lambda: is_active_list.append(True)
        )
        is_silent_writer.play()

        self.jump_to(duration_in_seconds)
        return bool(is_active_list)

    def is_pyo_object_silent(self, *args, **kwargs) -> bool:
        return not self.is_pyo_object_not_silent(*args, **kwargs)

    def setup_module_instance(
        self,
        module_instance: walkman.Module,
        module_container: typing.Optional[walkman.ModuleContainer] = None,
    ):
        if module_container is None:
            module_container = walkman.ModuleContainer({})
        module_container.prepare_module(module_instance)
        module_instance.initialise()
        for input_module_instance in module_instance.module_input_chain:
            input_module_instance.initialise()

    def setUp(self):
        self.server = pyo.Server(*self.server_args, **self.server_kwargs).boot().start()

    def tearDown(self):
        self.server.shutdown()


class ModuleTestCase(WalkmanTestCase, abc.ABC):
    @abc.abstractmethod
    def get_module_class(self) -> typing.Type[walkman.Module]:
        ...

    def get_module_instance(
        self,
        module_class: typing.Optional[typing.Type[walkman.Module]] = None,
        **kwargs
    ) -> walkman.Module:
        if module_class is None:
            module_class = self.get_module_class()
        module_instance = module_class(**kwargs)
        self.setup_module_instance(module_instance)
        return module_instance

    def play_module_instance(self, module_instance: walkman.Module, **kwargs):
        """Wrapper to play module instance and all dependencies"""
        for depending_module_instance in module_instance.module_chain:
            depending_module_instance.play(**kwargs)
        module_instance.play(**kwargs)

    def stop_module_instance(self, module_instance: walkman.Module, wait: float = 0):
        """Wrapper to stop module instance and all dependencies"""
        for depending_module_instance in module_instance.module_chain:
            depending_module_instance.stop(wait + module_instance.fade_out_duration)
        module_instance.stop(wait)

    def test_is_playing(self):
        module_instance = self.get_module_instance()
        self.assertFalse(module_instance.is_playing)
        self.assertFalse(module_instance.pyo_object.isPlaying())

        self.play_module_instance(
            module_instance,
        )
        self.assertTrue(module_instance.is_playing)
        self.assertTrue(module_instance.pyo_object.isPlaying())

        self.stop_module_instance(
            module_instance,
        )
        self.jump_to(module_instance.fade_out_duration + 0.1)
        self.assertFalse(module_instance.is_playing)
        self.assertFalse(module_instance.pyo_object.isPlaying())

    def test_play(self):
        module_instance = self.get_module_instance()
        self.assertTrue(self.is_pyo_object_silent(module_instance.pyo_object))

        self.play_module_instance(module_instance)
        self.assertFalse(self.is_pyo_object_silent(module_instance.pyo_object))

        for pyo_object in module_instance.internal_pyo_object_list:
            self.assertTrue(pyo_object.isPlaying())

    def test_play_with_delay(self):
        module_instance = self.get_module_instance()

        self.play_module_instance(module_instance, delay=4)
        # For 4 seconds it will still be silent
        self.assertFalse(module_instance.pyo_object.isPlaying())
        self.assertTrue(self.is_pyo_object_silent(module_instance.pyo_object, 4))

        self.assertTrue(module_instance.pyo_object.isPlaying())
        self.assertFalse(self.is_pyo_object_silent(module_instance.pyo_object, 4))

    def test_play_with_duration(self):
        module_instance = self.get_module_instance()

        self.play_module_instance(module_instance, duration=4)
        self.assertFalse(self.is_pyo_object_silent(module_instance.pyo_object, 4))

        self.assertTrue(self.is_pyo_object_silent(module_instance.pyo_object, 2))

    def test_stop(self):
        module_instance = self.get_module_instance()

        self.play_module_instance(module_instance)
        self.assertFalse(self.is_pyo_object_silent(module_instance.pyo_object))

        self.stop_module_instance(
            module_instance,
        )
        self.jump_to(module_instance.fade_out_duration)
        self.assertTrue(self.is_pyo_object_silent(module_instance.pyo_object))

        for pyo_object in module_instance.internal_pyo_object_list:
            # XXX: This doesn't work for mixer objects:
            #   SystemError: <class '_pyo.InputFader_base'> returned a result with an error set
            if not isinstance(pyo_object, pyo.Mixer):
                self.assertTrue(self.is_pyo_object_silent(pyo_object))
                self.assertFalse(pyo_object.isPlaying())

    def test_stop_with_wait(self):
        module_instance = self.get_module_instance()

        self.play_module_instance(module_instance)
        self.assertFalse(self.is_pyo_object_silent(module_instance.pyo_object))

        self.stop_module_instance(module_instance, wait=4)
        self.assertFalse(
            self.is_pyo_object_silent(
                module_instance.pyo_object, 4 + module_instance.fade_out_duration
            )
        )

        self.assertTrue(self.is_pyo_object_silent(module_instance.pyo_object))

    def test_fader(self):
        """Ensure pyo_object is a combination of _pyo_object and fader"""

        module_instance = self.get_module_instance(fade_in_duration=4)

        self.play_module_instance(module_instance)
        step_size = 0.1
        for _ in range(int(module_instance.fade_out_duration / step_size)):
            current_fader_sample = module_instance.fader.get()
            current_private_pyo_object_sample = module_instance._pyo_object.get()
            self.assertEqual(
                current_fader_sample * current_private_pyo_object_sample,
                module_instance.pyo_object.get(),
            )

    def test_fade_in(self):
        module_instance = self.get_module_instance(fade_in_duration=4)

        self.play_module_instance(module_instance)
        step_size = 0.1
        last_fader_sample = -0.001
        for _ in range(int(module_instance.fade_in_duration / step_size)):
            current_fader_sample = module_instance.fader.get()
            self.assertLess(last_fader_sample, current_fader_sample)
            last_fader_sample = current_fader_sample
            self.jump_to(step_size)

    def test_fade_out(self):
        module_instance = self.get_module_instance(fade_out_duration=10)

        self.play_module_instance(module_instance)
        self.assertFalse(self.is_pyo_object_silent(module_instance.pyo_object))

        self.stop_module_instance(
            module_instance,
        )
        step_size = 0.1
        last_fader_sample = 1.001
        for _ in range(int(module_instance.fade_out_duration / step_size)):
            current_fader_sample = module_instance.fader.get()
            self.assertGreater(last_fader_sample, current_fader_sample)
            self.assertTrue(module_instance.pyo_object.isPlaying())
            self.assertFalse(
                self.is_pyo_object_silent(module_instance.pyo_object, step_size)
            )
            last_fader_sample = current_fader_sample

        self.assertFalse(module_instance.pyo_object.isPlaying())
        self.assertTrue(self.is_pyo_object_silent(module_instance.pyo_object, 2))

    def test_initialise_basic(self):
        """Ensure there are no errors when calling initialise method."""
        module_instance = self.get_module_instance(fade_out_duration=10)
        module_instance.initialise()

    def test_send_to_physical_output(self):
        module_instance = self.get_module_instance(send_to_physical_output=True)
        self.play_module_instance(module_instance)

        self.assertTrue(module_instance.pyo_object.isOutputting())

    def test_do_not_send_to_physical_output(self):
        module_instance = self.get_module_instance(send_to_physical_output=False)
        self.play_module_instance(module_instance)

        self.assertFalse(module_instance.pyo_object.isOutputting())

    def test_no_module_input_argument_in_private_initialse(self):
        # If there are any arguments with equal name as
        # names of input modules, the passed arguments of a user
        # will never reach the _initialise method because
        # of walkmans syntactic sugar (defined in initialise method
        # of base.Module). Therefore a module class should never
        # add any argument like this to its _initialise method.
        module_class = self.get_module_class()
        for parameter in inspect.signature(module_class._initialise).parameters:
            self.assertTrue(parameter not in module_class.default_module_input_dict)



from __future__ import annotations

import functools
import time

import pyo

import walkman

__all__ = (
    "PlayMixin",
    "JumpToMixin",
    "CloseMixin",
    "NamedMixin",
    "PyoObjectMixin",
    "DecibelMixin",
)


class PlayMixin(object):
    @property
    def is_playing(self) -> bool:
        try:
            return self._is_playing
        except AttributeError:
            return False

    def _play(self, duration: float = 0, delay: float = 0):
        ...

    def _stop(self, wait: float = 0):
        ...

    def play(self, duration: float = 0, delay: float = 0) -> PlayMixin:
        if not self.is_playing:
            self._play(duration=duration, delay=delay)
            self._is_playing = True
        return self

    def stop(self, wait: float = 0) -> PlayMixin:
        if self.is_playing:
            self._stop(wait=wait)
            self._is_playing = False
        return self


class JumpToMixin(object):
    def jump_to(self, time_in_seconds: float):
        ...


class TrackedPlayMixin(PlayMixin, JumpToMixin):
    @property
    def play_time(self) -> float:
        """Get time when 'play' method has been called"""
        return self._play_time

    @property
    def current_time(self) -> float:
        """Get current position of object"""
        return self._current_time

    def _jump_to(self, time_in_seconds: float):
        ...

    def _play(self, duration: float = 0, delay: float = 0):
        self._jump_to(self._current_time)
        self._play_time = time.time()

    def _stop(self, wait: float = 0):
        self._current_time = min(
            ((time.time() - self.play_time) + self.play_time, self.duration)
        )

    def jump_to(self, time_in_seconds: float):
        self._current_time = time_in_seconds
        self._jump_to(time_in_seconds)


class CloseMixin(object):
    def close(self) -> CloseMixin:
        return self


class NamedMixin(object):
    @classmethod
    def get_class_name(cls) -> str:
        # class name
        return walkman.utilities.camel_case_to_snake_case(cls.__name__)

    def get_instance_name(self) -> str:
        try:
            return self._instance_name
        except AttributeError:
            self._instance_name = f"{self.get_class_name()}-{id(self)}"
            return self.get_instance_name()

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.get_instance_name()})"

    def __hash__(self) -> int:
        return hash(self.get_instance_name())


class PyoObjectMixin(object):
    @functools.cached_property
    def pyo_object(self) -> pyo.PyoObject:
        return pyo.Sig(0)


class DecibelMixin(object):
    @property
    def decibel(self) -> float:
        try:
            return self._decibel.getValue()
        except AttributeError:
            return -120

    @decibel.setter
    def decibel(self, value: float) -> float:
        try:
            self._decibel.setValue(value)
        except AttributeError:
            pass



"""Setup user interface for walkman"""

import abc
import dataclasses
import datetime
import functools
import operator
import time
import typing
import warnings

import PySimpleGUI as sg

import walkman


@dataclasses.dataclass(frozen=True)
class UIElement(abc.ABC):
    backend: walkman.Backend
    key_tuple: typing.Optional[str] = None
    keyboard_key_tuple: typing.Optional[str] = None

    @property
    def gui_element(self) -> typing.Union[list, sg.Element]:
        return []

    @abc.abstractmethod
    def handle_event(self, event: str, value_dict: dict):
        ...

    @functools.cached_property
    def event_tuple(self) -> typing.Tuple[str, ...]:
        event_list = []
        if self.key_tuple:
            event_list.extend(self.key_tuple)
        if self.keyboard_key_tuple:
            event_list.extend(self.keyboard_key_tuple)
        return tuple(event_list)

    def tick(self, value_dict: dict):
        # Called in every loop part
        pass


@dataclasses.dataclass(frozen=True)
class SimpleUIElement(UIElement):
    element_args: tuple = tuple([])
    element_kwargs: dict = dataclasses.field(default_factory=lambda: {})
    pysimple_gui_class: typing.Optional[typing.Type[sg.Element]] = None

    @functools.cached_property
    def gui_element(self) -> typing.Union[list, sg.Element]:
        if self.pysimple_gui_class:
            return self.get_element_instance()
        return None

    def get_element_instance(self) -> sg.Element:
        assert self.pysimple_gui_class is not None
        return self.pysimple_gui_class(*self.element_args, **self.element_kwargs)

    def handle_event(self, _: str, __: dict):
        pass


class NestedUIElement(UIElement):
    def __init__(
        self,
        backend,
        ui_element_sequence: typing.Sequence[UIElement],
        **kwargs,
    ):
        super().__init__(backend=backend, **kwargs)
        self.ui_element_tuple = tuple(ui_element_sequence)

        event_to_ui_element_dict = {}
        for ui_element in self.ui_element_tuple:
            for event in ui_element.event_tuple:
                if event not in event_to_ui_element_dict:
                    event_to_ui_element_dict.update({event: ui_element})
                else:
                    raise DuplicateEventError(event)
        self.event_to_ui_element_dict = event_to_ui_element_dict

    @functools.cached_property
    def event_tuple(self) -> typing.Tuple[str, ...]:
        return (
            functools.reduce(
                operator.add,
                (ui_element.event_tuple for ui_element in self.ui_element_tuple),
            )
            + super().event_tuple
        )

    @property
    def simple_ui_element_tuple(self) -> typing.Tuple[SimpleUIElement, ...]:
        simple_ui_element_list = []
        for ui_element in self.ui_element_tuple:
            try:
                simple_ui_element_tuple = ui_element.simple_ui_element_tuple
            except AttributeError:
                simple_ui_element_list.append(ui_element)
            else:
                simple_ui_element_list.extend(simple_ui_element_tuple)
        return tuple(simple_ui_element_list)

    def handle_event(self, event: str, value_dict: dict):
        try:
            ui_element = self.event_to_ui_element_dict[event]
        except KeyError:
            pass
        else:
            return ui_element.handle_event(event, value_dict)

    def tick(self, value_dict: dict):
        for ui_element in self.ui_element_tuple:
            ui_element.tick(value_dict)

    @functools.cached_property
    def gui_element(self) -> list:
        return [
            ui_element.gui_element
            for ui_element in self.ui_element_tuple
            if ui_element.gui_element
        ]


class StopWatch(UIElement):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, key_tuple=("stop_watch",), **kwargs)
        self._start_time = time.time()
        self._last_time: typing.Optional[float] = None
        self._display_time: float = 0
        self.stop()

    @staticmethod
    def format_time(time_to_format: float) -> str:
        _, minutes, seconds = str(datetime.timedelta(seconds=time_to_format)).split(":")
        return f"{minutes}:{seconds[:2]}"

    @property
    def duration(self) -> float:
        return self.backend.cue_manager.current_cue.duration

    @property
    def formatted_duration(self) -> str:
        return StopWatch.format_time(self.duration)

    @property
    def current_time(self) -> float:
        if self.duration > 0:
            return time.time() - self._start_time
        else:
            return 0

    @property
    def current_time_formatted(self) -> str:
        return StopWatch.format_time(self.current_time)

    @property
    def display_time(self) -> float:
        return self._display_time

    @functools.cached_property
    def gui_element(self) -> typing.Union[list, sg.Element]:
        return sg.Text(self._get_update_string())

    def start(self):
        if self._last_time is not None:
            self.set_to(self._last_time)
            self._last_time = None

    def stop(self):
        self._last_time = self.current_time

    def reset(self):
        self._start_time = time.time()
        self.update()

    def set_to(self, time_in_seconds: float):
        self._start_time = time.time() - time_in_seconds

    def _get_update_string(self) -> str:
        return f"{self.current_time_formatted} // {self.formatted_duration}"

    def update(self):
        if (current_time := self.current_time) <= self.duration:
            self._display_time = current_time
            self.gui_element.update(self._get_update_string())

    def handle_event(self, _: str, __: dict):
        self.update()

    def tick(self, value_dict: dict):
        self.update()



class Button(SimpleUIElement):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, pysimple_gui_class=sg.Button, **kwargs)
        self.element_kwargs.update({"key": self.key_tuple[0]})


class TooLargeTimeWarning(Warning):
    def __init__(self, time: float):
        super().__init__(
            f"Can't jump to given time: {time} seconds longer than cue duration."
        )


class JumpToTimeButton(Button):
    def __init__(
        self,
        stop_watch: StopWatch,
        jump_to_time_input_minutes_key: str,
        jump_to_time_input_seconds_key: str,
        *args,
        **kwargs,
    ):
        self.stop_watch = stop_watch
        self.key = "jump_to_time_button"
        self.jump_to_time_input_minutes_key = jump_to_time_input_minutes_key
        self.jump_to_time_input_seconds_key = jump_to_time_input_seconds_key
        super().__init__(
            *args,
            key_tuple=(self.key,),
            element_kwargs={"button_text": "JUMP TO"},
            **kwargs,
        )

    def handle_event(self, _: str, value_dict: dict):
        minutes = value_dict[self.jump_to_time_input_minutes_key]
        seconds = value_dict[self.jump_to_time_input_seconds_key]
        time = (int(minutes) * 60) + float(seconds)
        if time <= (cue_duration := self.backend.cue_manager.current_cue.duration):
            for module in self.backend.module_container.module_tuple:
                module.jump_to(time)
            self.stop_watch.set_to(time)
            if not self.backend.audio_host.is_playing:
                self.stop_watch.stop()
            self.stop_watch.update()
        else:
            warnings.warn(TooLargeTimeWarning(time - cue_duration))


class FrozenText(SimpleUIElement):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, pysimple_gui_class=sg.Text, **kwargs)


class Popup(SimpleUIElement):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, pysimple_gui_class=sg.Popup, **kwargs)
        self.popup = None

    @functools.cached_property
    def gui_element(self) -> typing.Union[list, sg.Element]:
        return []

    def handle_event(self, _: str, value_dict: dict):
        self.popup = self.get_element_instance()


class Window(SimpleUIElement):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, pysimple_gui_class=sg.Window, **kwargs)


class ModalWindow(Window):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.element_kwargs.update({"modal": True})


class TitleBar(SimpleUIElement):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, pysimple_gui_class=sg.Titlebar, **kwargs)


class Slider(SimpleUIElement):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, pysimple_gui_class=sg.Slider, **kwargs)


class Title(TitleBar):
    def __init__(self, backend: walkman.Backend, *args, **kwargs):
        element_kwargs = {
            "title": f"{walkman.constants.NAME}: {backend.name}",
            "icon": walkman.constants.ICON,
        }
        super().__init__(backend, *args, element_kwargs=element_kwargs, **kwargs)


class Menu(SimpleUIElement):
    sg.MENU_SHORTCUT_CHARACTER = "&"

    # Please see https://github.com/PySimpleGUI/PySimpleGUI/issues/4072#issuecomment-803355769
    @staticmethod
    def Menubar(
        menu_definition,
        text_color: str = "black",
        background_color: str = "white",
        pad=(0, 0),
    ):
        """
        A User Defined element that simulates a Menu element by using ButtonMenu elements

        :param menu_definition: A standard PySimpleGUI menu definition
        :type menu_definition: List[List[Tuple[str, List[str]]]
        :param text_color: color for the menubar's text
        :type text_color:
        :param background_color: color for the menubar's background
        :type background_color:
        :param pad: Amount of padding around each menu entry
        :type pad:
        :return: A column element that has a row of ButtonMenu buttons
        :rtype: sg.Column
        """
        row = []
        for menu in menu_definition:
            text = menu[0]
            if text.__contains__(sg.MENU_SHORTCUT_CHARACTER):
                text = text.replace(sg.MENU_SHORTCUT_CHARACTER, "")
            if text.startswith(sg.MENU_DISABLED_CHARACTER):
                disabled = True
                text = text[len(sg.MENU_DISABLED_CHARACTER) :]
            else:
                disabled = False
            row += [
                sg.ButtonMenu(
                    text,
                    menu,
                    border_width=0,
                    button_color=f"{text_color} on {background_color}",
                    key=text,
                    pad=pad,
                    disabled=disabled,
                )
            ]

        return sg.Column(
            [row], background_color=background_color, pad=(0, 0), expand_x=True
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, pysimple_gui_class=self.Menubar, **kwargs)


HELP_KEY = "Help"
CHANNEL_TEST_KEY = "Channel test"
ROTATION_CHANNEL_TEST_KEY = "Launch rotation test"


class WalkmanMenu(Menu):
    def __init__(self, backend: walkman.Backend, *args, **kwargs):
        super().__init__(
            backend,
            *args,
            element_args=(
                [
                    [
                        f"&{CHANNEL_TEST_KEY}",
                        [ROTATION_CHANNEL_TEST_KEY, "Launch individual channel test"],
                    ],
                    [f"&{HELP_KEY}", "&About..."],
                ],
            ),
            **kwargs,
        )


class Logging(UIElement):
    def handle_event(self, _: str, __: dict):
        pass


class AboutText(Popup):
    def __init__(self, backend: walkman.Backend, *args, **kwargs):
        element_args = (
            f"Welcome to {walkman.constants.NAME}. "
            "This is a software for audio cue control. "
            "Please consult the github page "
            "for futher information at "
            "https://github.com/audiowalkman/walkman.",
        )
        super().__init__(
            backend,
            *args,
            element_args=element_args,
            key_tuple=(HELP_KEY,),
            **kwargs,
        )


class JumpToTimeInput(UIElement):
    @functools.cached_property
    def gui_element(self) -> typing.Union[list, sg.Element]:
        return sg.InputText(
            key=self.key_tuple[0], default_text="0", enable_events=False, size=(4, 1)
        )

    def handle_event(self, _: str, __: dict):
        pass


class JumpToTimeInputMinutes(JumpToTimeInput):
    def __init__(self, *args, **kwargs):
        self.key = "jump_to_time_input_minutes"
        super().__init__(*args, key_tuple=(self.key,), **kwargs)


class JumpToTimeInputSeconds(JumpToTimeInput):
    def __init__(self, *args, **kwargs):
        self.key = "jump_to_time_input_seconds"
        super().__init__(*args, key_tuple=(self.key,), **kwargs)


class StartStopButton(Button):
    def __init__(
        self,
        *args,
        key_tuple=("start_stop",),
        element_kwargs={"button_text": "START // STOP"},
        **kwargs,
    ):
        super().__init__(
            *args,
            key_tuple=key_tuple,
            element_kwargs=element_kwargs,
            keyboard_key_tuple=(" ", "space:65"),
            **kwargs,
        )
        self.is_playing = 0

    def play(self):
        self.gui_element.update(button_color="red")

    def stop(self):
        self.gui_element.update(button_color="white")

    def handle_event(self, _: str, __: dict):
        if self.is_playing:
            self.stop()
        else:
            self.play()
        self.is_playing = not self.is_playing


class WalkmanStartStopButton(StartStopButton):
    def __init__(self, stop_watch: StopWatch, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
        )
        self.stop_watch = stop_watch

    def play(self):
        super().play()
        self.gui_element.update(button_color="red")
        self.backend.cue_manager.current_cue.play()
        self.stop_watch.start()

    def stop(self):
        super().stop()
        self.backend.cue_manager.current_cue.stop()
        self.stop_watch.stop()

    def tick(self, value_dict: dict):
        if self.is_playing:
            self.stop_watch.tick(value_dict)


class AudioTestStartStopButton(StartStopButton):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
        )
        self.audio_test = None

    def play(self):
        super().play()
        if self.audio_test:
            self.audio_test.play()

    def stop(self):
        super().stop()
        if self.audio_test:
            self.audio_test.stop()


class VolumeSlider(Slider):
    def __init__(
        self,
        *args,
        default_value: typing.Optional[float] = -12,
        key_suffix: str = "",
        volume_range: typing.Tuple[float, float] = (-120, 12),
        resolution: float = 0.25,
        audio_object_with_decibel: typing.Optional[walkman.DecibelMixin] = None,
        orientation: str = "horizontal",
        tooltip: typing.Optional[str] = None,
        **kwargs,
    ):
        self.key = f"volume_{key_suffix}"
        super().__init__(
            *args,
            element_kwargs={
                "default_value": default_value,
                "key": self.key,
                "range": volume_range,
                "resolution": resolution,
                "orientation": orientation,
                "enable_events": True,
                "tooltip": tooltip,
            },
            **kwargs,
        )
        self.audio_object_with_decibel = audio_object_with_decibel

    def set_decibel(self, value_dict: dict):
        if (
            self.audio_object_with_decibel
            and value_dict
            and (decibel := value_dict.get(self.key, None))
        ):
            self.audio_object_with_decibel.decibel = decibel

    def handle_event(self, event: str, value_dict: dict):
        self.set_decibel(value_dict)

    def tick(self, value_dict: dict):
        self.set_decibel(value_dict)


class SelectCueMenu(UIElement):
    def __init__(self, stop_watch: StopWatch, *args, **kwargs):
        self.stop_watch = stop_watch
        self.combo_key = "select_cue"

        self.left_key = "Left:113"
        self.right_key = "Right:114"

        super().__init__(
            *args,
            key_tuple=(self.combo_key,),
            keyboard_key_tuple=(self.left_key, self.right_key),
            **kwargs,
        )

        self.value_tuple = self.backend.cue_manager.cue_name_tuple
        self.value_count = len(self.value_tuple)

    @functools.cached_property
    def gui_element(self) -> typing.Union[list, sg.Element]:
        return sg.Combo(
            self.value_tuple,
            default_value=self.value_tuple[0],
            readonly=True,
            enable_events=True,
            key=self.combo_key,
        )

    def handle_event(self, event: str, value_dict: dict):
        cue_name = value_dict[self.combo_key]
        if event == self.combo_key:
            new_cue_index = self.backend.cue_manager.cue_name_tuple.index(cue_name)
        else:
            cue_index = self.value_tuple.index(cue_name)
            # When current duration != 0 and one left key: jump to beginning
            # of current cue.
            if event == self.left_key and int(self.stop_watch.display_time) != 0:
                new_cue_index = cue_index
            else:
                new_cue_index = (
                    cue_index + {self.left_key: -1, self.right_key: 1}[event]
                ) % self.value_count
            cue_name = self.value_tuple[new_cue_index]
            self.gui_element.update(value=cue_name)

        self.backend.cue_manager.jump_to_cue(new_cue_index)
        self.stop_watch.reset()
        self.stop_watch.stop()


class DuplicateEventError(Exception):
    def __init__(self, event: str):
        super().__init__(f"Event '{event}' has been defined twice!")


class JumpToTimeControl(NestedUIElement):
    def __init__(
        self,
        stop_watch: StopWatch,
        backend,
    ):
        self.jump_to_time_input_minutes = JumpToTimeInputMinutes(backend)
        self.jump_to_time_input_seconds = JumpToTimeInputSeconds(backend)
        self.jump_to_time_button = JumpToTimeButton(
            stop_watch,
            self.jump_to_time_input_minutes.key,
            self.jump_to_time_input_seconds.key,
            backend,
        )

        ui_element_sequence = (
            self.jump_to_time_button,
            self.jump_to_time_input_minutes,
            FrozenText(backend, element_kwargs={"text": "MIN"}),
            self.jump_to_time_input_seconds,
            FrozenText(backend, element_kwargs={"text": "SEC"}),
        )

        super().__init__(backend, ui_element_sequence)


class Transport(NestedUIElement):
    def __init__(
        self,
        backend,
    ):
        self.stop_watch = StopWatch(backend)
        self.start_stop_button = WalkmanStartStopButton(self.stop_watch, backend)
        self.select_cue_menu = SelectCueMenu(self.stop_watch, backend)

        ui_element_sequence = (
            self.start_stop_button,
            self.select_cue_menu,
            self.stop_watch,
        )

        super().__init__(backend, ui_element_sequence)

    def tick(self, value_dict: dict):
        for ui_element in self.ui_element_tuple:
            # Stop watch tick is handled in StartStopButton
            if ui_element != self.stop_watch:
                ui_element.tick(value_dict)


class CueControl(NestedUIElement):
    def __init__(
        self,
        backend,
    ):
        self.transport = Transport(backend)
        self.jump_to_time_control = JumpToTimeControl(
            self.transport.stop_watch, backend
        )
        self.volume_slider = VolumeSlider(
            backend,
            key_suffix="master",
            audio_object_with_decibel=backend.audio_host,
            tooltip="Set volume of master output",
        )

        ui_element_sequence = (
            self.transport,
            self.jump_to_time_control,
            self.volume_slider,
        )

        super().__init__(backend, ui_element_sequence)


class Walkman(NestedUIElement):
    def __init__(
        self,
        backend,
    ):
        self.menu = WalkmanMenu(backend)
        self.about_text = AboutText(backend)
        self.audio_rotation_test = AudioRotationTest(backend)
        self.cue_control = CueControl(backend)

        ui_element_sequence = (
            self.cue_control,
            self.menu,
            self.about_text,
            self.audio_rotation_test,
        )

        super().__init__(backend, ui_element_sequence)

    @functools.cached_property
    def gui_element(self) -> list:
        return [
            [self.menu.gui_element],
            self.cue_control.gui_element,
            self.about_text.gui_element,
        ]


class NestedWindow(NestedUIElement):
    def __init__(
        self,
        backend: walkman.Backend,
        ui_element_sequence: typing.Sequence[UIElement],
        window_class: typing.Type[Window] = Window,
        window_name: str = walkman.constants.NAME,
        window_kwargs: typing.Dict[str, typing.Any] = {},
        **kwargs,
    ):
        window_kwargs.update(
            {
                "resizable": True,
                "scaling": 3,
                "return_keyboard_events": True,
            }
        )
        super().__init__(backend, ui_element_sequence, **kwargs)
        self.window_name = window_name
        self.window_kwargs = window_kwargs
        self.window_class = window_class

    def set_window(self):
        # Initialisation of windows have to be dynamic (so
        # when closing and reopening windows we don't use the same
        # layout, which is prohibited).
        self.window = self.window_class(
            self.backend,
            element_args=(self.window_name, self.gui_element),
            element_kwargs=self.window_kwargs,
        )

    def before_loop(self):
        self.set_window()
        self.pysimple_gui_window = self.window.get_element_instance()
        self.pysimple_gui_window.finalize()

    def after_loop(self):
        self.pysimple_gui_window.close()
        del self.pysimple_gui_window
        for ui_element in self.simple_ui_element_tuple + (self,):
            try:
                del ui_element.gui_element
            except AttributeError:
                pass

    def read(self, pysimple_gui_window: sg.Window) -> bool:
        shall_break = False
        event, value_dict = pysimple_gui_window.read(timeout=1000)

        # See if user wants to quit or window was closed
        if event == sg.WINDOW_CLOSED or event == "Quit":
            shall_break = True

        # Ignore timeout events
        if event != sg.TIMEOUT_EVENT:
            walkman.constants.LOGGER.debug(
                f"Catched event       = '{event}' \n"
                f"with value_dict     = '{value_dict}'\n."
            )
            try:
                self.handle_event(event, value_dict)
            except Exception:
                walkman.constants.LOGGER.exception(
                    f"Catched exception when handling event '{event}'"
                    f"with value_dict = '{value_dict}': "
                )

        try:
            self.tick(value_dict)
        except Exception:
            walkman.constants.LOGGER.exception(
                "Catched exception when running tick method: "
            )

        return shall_break

    def loop(self):
        self.before_loop()
        while True:
            shall_break = self.read(self.pysimple_gui_window)
            if shall_break:
                break
        self.after_loop()


class AudioRotationTest(NestedWindow):
    def __init__(self, backend: walkman.Backend, *args, **kwargs):
        self.start_stop_button = AudioTestStartStopButton(backend)
        self.volume_slider = VolumeSlider(backend, key_suffix="rotation_test")
        ui_element_tuple = (self.start_stop_button, self.volume_slider)
        super().__init__(
            backend,
            ui_element_tuple,
            window_class=ModalWindow,
            key_tuple=(CHANNEL_TEST_KEY,),
            window_kwargs={"finalize": True},
            **kwargs,
        )
        self.event_tuple = self.key_tuple

    def before_loop(self):
        super().before_loop()
        self.audio_rotation_test = self.backend.get_audio_test(
            audio_test_class=walkman.tests.AudioRotationTest
        )
        self.start_stop_button.audio_test = self.audio_rotation_test
        self.volume_slider.audio_object_with_decibel = self.audio_rotation_test

    def after_loop(self):
        super().after_loop()
        self.audio_rotation_test.close()
        self.start_stop_button.audio_rotation_test = None
        del self.audio_rotation_test

    def handle_event(self, event: str, value_dict: dict):
        if (
            value_dict
            and value_dict.get(CHANNEL_TEST_KEY, None) == ROTATION_CHANNEL_TEST_KEY
        ):
            self.loop()
        else:
            super().handle_event(event, value_dict)

    @property
    def gui_element(self) -> list:
        return [
            [self.start_stop_button.gui_element],
            [self.volume_slider.gui_element],
        ]


class GUI(NestedWindow):
    def __init__(
        self,
        backend: walkman.Backend,
        *args,
        theme: str = "DarkBlack",
        **kwargs,
    ):
        sg.theme(theme)
        super().__init__(
            backend,
            *args,
            window_name=backend.name,
            window_kwargs={"icon": walkman.constants.ICON},
            **kwargs,
        )

    def before_loop(self):
        super().before_loop()
        self.backend.audio_host.play()

    def after_loop(self):
        super().after_loop()
        self.backend.cue_manager.current_cue.stop()
        self.backend.module_container.close()
        self.backend.audio_host.close()


UI_CLASS_TUPLE = (Walkman,)
"""All used UI classes"""


def backend_to_gui(backend: walkman.Backend) -> GUI:
    gui = GUI(backend, [ui_class(backend=backend) for ui_class in UI_CLASS_TUPLE])
    return gui



import typing
import warnings

import jinja2
import tomli

import walkman

CONFIGURE_KEY = "configure"
CONFIGURE_NAME_KEY = "name"
CONFIGURE_AUDIO_KEY = "audio"
CONFIGURE_AUDIO_CHANNEL_COUNT_KEY = "channel_count"
CONFIGURE_MODULE_KEY = "module"

CUE_KEY = "cue"


def pop_from_dict(dict_to_pop_from: dict, name: str, fallback_value: typing.Any = None):
    try:
        value = dict_to_pop_from[name]
    except KeyError:
        value = fallback_value
    else:
        del dict_to_pop_from[name]
    return value


class UnusedSpecificationWarning(Warning):
    pass


def warn_not_used_configuration_content(toml_block: dict, block_name: str):
    if toml_block:
        formatted_toml_block = "\t>>> ".join(str(toml_block).splitlines(True))
        warnings.warn(
            "WALKMAN ignored the following invalid specifications "
            f"in '{block_name.upper()}':\n{formatted_toml_block}",
            UnusedSpecificationWarning,
        )


def add_replication_instance(
    replication_key: int,
    module_name: str,
    replication_configuration: dict,
    configuration: dict,
):
    if replication_key in replication_configuration:
        warnings.warn(
            "Found repeated replication_key '{replication_key}' "
            f"in module '{module_name}'!"
        )
    replication_configuration.update({replication_key: configuration})


def configure_module_block_and_audio_object_to_module_container(
    configure_module_block: dict,
) -> walkman.ModuleContainer:
    module_name_to_replication_configuration_dict = {}
    for module_name, replication_configuration_block in configure_module_block.items():
        replication_configuration = {}
        for replication_key, configuration in replication_configuration_block.items():
            add_replication_instance(
                replication_key, module_name, replication_configuration, configuration
            )
        module_name_to_replication_configuration_dict.update(
            {module_name: replication_configuration}
        )
    return walkman.ModuleContainer.from_module_configuration(
        module_name_to_replication_configuration_dict,
    )


def configure_block_to_global_state_object_tuple(
    configure_block: dict,
) -> typing.Tuple[str, walkman.AudioHost, walkman.ModuleContainer]:
    name = pop_from_dict(configure_block, CONFIGURE_NAME_KEY, "Project")
    audio_block = pop_from_dict(configure_block, CONFIGURE_AUDIO_KEY, {})
    module_block = pop_from_dict(configure_block, CONFIGURE_MODULE_KEY, {})

    warn_not_used_configuration_content(configure_block, "configure")

    audio_host = walkman.AudioHost(**audio_block)
    module_container = configure_module_block_and_audio_object_to_module_container(
        module_block
    )

    return (name, audio_host, module_container)


def cue_block_and_module_container_to_cue_manager(
    cue_block: dict, module_container: walkman.ModuleContainer
) -> walkman.CueManager:
    cue_list = []
    for cue_name, cue_kwargs in cue_block.items():
        module_name_to_replication_configuration = {}
        for module_name, replication_configuration_block in cue_kwargs.items():
            replication_configuration = {}
            for (
                replication_key,
                module_kwargs,
            ) in replication_configuration_block.items():
                add_replication_instance(
                    replication_key,
                    module_name,
                    replication_configuration,
                    module_kwargs,
                )
            module_name_to_replication_configuration.update(
                {module_name: replication_configuration}
            )

        cue = walkman.Cue(
            module_container, cue_name, **module_name_to_replication_configuration
        )
        cue_list.append(cue)

    cue_manager = walkman.CueManager(tuple(cue_list))
    return cue_manager


def toml_str_to_backend(
    toml_str: str = "",
) -> walkman.Backend:
    toml_dictionary = tomli.loads(toml_str)

    configure_block = pop_from_dict(toml_dictionary, CONFIGURE_KEY, {})
    cue_block = pop_from_dict(toml_dictionary, CUE_KEY, {})
    warn_not_used_configuration_content(toml_dictionary, "global")

    global_state_object_tuple = configure_block_to_global_state_object_tuple(
        configure_block
    )
    cue_manager = cue_block_and_module_container_to_cue_manager(
        cue_block, global_state_object_tuple[-1]
    )

    backend = walkman.Backend(*global_state_object_tuple, cue_manager)
    return backend


def toml_file_path_to_backend(toml_file_path: str) -> walkman.Backend:
    with open(toml_file_path, "r") as toml_file:
        toml_str = toml_file.read()
    return toml_str_to_backend(toml_str)


def jinja2_file_path_to_backend(jinja2_file_path: str) -> walkman.Backend:
    environment = jinja2.Environment(loader=jinja2.FileSystemLoader("./"))
    template = environment.get_template(jinja2_file_path)
    toml_str = template.render()
    toml_file_path = f".{jinja2_file_path.split('.')[0]}.toml"
    with open(toml_file_path, "w") as toml_file:
        toml_file.write(toml_str)
    walkman.constants.LOGGER.info(
        "WALKMAN converted jinja2 template to toml file"
        f"\n\nPlease check '{toml_file_path}' "
        "for converted file.\n"
    )
    return toml_str_to_backend(toml_str)
